#!/usr/bin/python
#############################################################
#############################################################
# Auto Drive
# Author: Can Kosar
#############################################################

#Standard libraries
import sys
import os
import time
import math
import numpy as np
from scipy import integrate

#Pickle
import pickle
from signal import signal, SIGPIPE, SIG_DFL
signal(SIGPIPE,SIG_DFL) 
#Custom libraries
sys.path.append("../lib/")
import l_localize
import l_drive
import l_config
import l_us

# sgn=drive.move_fw(40,0)
# #sgn=[1,1]
# state.set_odo_signs(sgn)


class sequence():
    
    def __init__(self):
        #State
        self.state=l_localize.localize_asynch()
        self.state.start()
        self.state.set_odo_signs([1,1])#:s
        #Ultrasonic transducer
        self.us=l_us.us_asynch()
        self.us.start()
        self.us_data=[]
        #Drive
        self.drive=l_drive.drive()
        self.config=l_config.config()
        self.fo_time=0.5
        self.duty_turn=70
        self.duty_straight=50
        self.balance_calibration=12
        self.eps_lin=0.005
        self.eps_turn=math.radians(1)
        time.sleep(0.1)
    
    
    def execute_sequence(self,seq):
        
        self.KOS_ref=np.matrix([0,0,0]).transpose()
        
        for p in seq:
            if p[0]=="fw":
                transl=np.matrix([p[2],0,0]).transpose()                
            elif p[0]=="tr":
                transl=np.matrix([0,0,p[2]]).transpose()              
                
            nv=self.local_to_global(self.KOS_ref[:,-1],transl)
            self.KOS_ref=np.concatenate((self.KOS_ref,nv),axis=1)
        
        
        for p in range(len(seq)):
            #print p
            
            if seq[p][0]=="fw":
                self.fw(seq[p],self.KOS_ref[:,p])
            elif seq[p][0]=="tr":
                self.tr(seq[p],self.KOS_ref[:,p])
                    
    def local_to_global(self,ref,vector):
        
        #print ref
        alpha=ref[2]
#         print d_alpha_gl
        #Transformation-matrix (Local to global)
        T= np.matrix([[math.cos(alpha), -math.sin(alpha),0],
                      [math.sin(alpha), math.cos(alpha),0],
                      [0,0,1]])
        #Displacement in the global system
        #du_glob=T*cos        
        return ref+T*vector
                    
                    
    def global_to_local(self,ref,vector,mode=False):

        alpha=-ref[2]
        #Transformation-matrix (Global to local)
        T= np.matrix([[math.cos(alpha), -math.sin(alpha),0],
                      [math.sin(alpha), math.cos(alpha),0],
                      [0,0,1]])
        #Local system
        if mode=="rotate":
            return T*(vector) 
        else:
            return T*(vector-ref)            
                    
                    
    def co_simulation_straight(self,y_0):
        #global sim_a_t_s,sim_u_s,sim_u_dot_s
        
        delay=0.001
        
        def ode_equ(y, t):
            m=self.config.hardware.mass
            r=self.config.hardware.r_robot
            d=12
            return [y[1], -d/m*y[1]]
        
        
        #Delay
        a_t_delay=[0,delay]
        sim_u_dot_delay=[y_0[1],y_0[1]]    
        sim_u_delay=[y_0[0],y_0[0]+delay*y_0[1]]
        
        #Damping
        a_t = np.arange(0+delay, 1.0+delay, 0.2)
        y_after_delay=[sim_u_delay[-1],sim_u_dot_delay[-1]]    
        asol = integrate.odeint(ode_equ, y_after_delay, a_t)
        
        #print a_t_delay
        #print sim_u_delay
        #print sim_u_dot_delay
        #sys.exit()
        sim_a_t_s=np.concatenate((a_t_delay,a_t)).tolist()
        sim_u_s=np.concatenate((sim_u_delay,asol[:,0])).tolist()
        sim_u_dot_s=np.concatenate((sim_u_dot_delay,asol[:,1])).tolist()
    
        return sim_u_s[-1]
                    
    def co_simulation_turn(self,y_0):
        #print "csy0", y_0
        delay=0.001
        
        def ode_equ(y, t):
            m=self.config.hardware.mass
            r=self.config.hardware.r_robot
            Iz=(m*(r**2))/2
            d=0.13
            return [y[1], -d/Iz*y[1]]
        
        
        #Delay
        a_t_delay=[0,delay]
        sim_u_dot_delay=[y_0[1],y_0[1]]    
        sim_u_delay=[y_0[0],y_0[0]+delay*y_0[1]]
        
        #Damping
        a_t = np.arange(0+delay, 1.0+delay, 0.2)
        y_after_delay=[sim_u_delay[-1],sim_u_dot_delay[-1]]    
        asol = integrate.odeint(ode_equ, y_after_delay, a_t)

        sim_a_t=np.concatenate((a_t_delay,a_t)).tolist()
        sim_u=np.concatenate((sim_u_delay,asol[:,0])).tolist()
        sim_u_dot=np.concatenate((sim_u_dot_delay,asol[:,1])).tolist()
    
        return sim_u[-1]                
    
    def fw(self,p,ref):
        
        fo_start=False

        while True:
            #Get state / Position /Velocity in global and local coordinates
            #print self.state.t_label
            t=self.state.t[0]
            y=np.matrix(self.state.y[0:3]).transpose()
            yd=np.matrix(self.state.y[3:6]).transpose()
            yl=self.global_to_local(ref,y)
            ydl=self.global_to_local(ref,yd,"rotate")
            
            #Ultrasonic
            us_arr=[float(i) for j in self.us.distance for i in j]
            self.us_data.append({"state":y,"us":us_arr})
            
            #print "ydl",ydl
            #Controlling
            if True:
                
                e=yl
                ed=ydl
                
                #State space
                ss=np.concatenate((e,ed),axis=0).transpose()
                #Parameter matrix    
                A=np.matrix([0,-1000,-150,0,-1,-0.5]).transpose()
                #Set matrix
                #print l_misc.plots().print_matrix(ss)
                #print l_misc.plots().print_matrix(A)
                sm=ss*A       
                
                #PD-Controller
                #print l_misc.plots().print_matrix(e)        
                new_balance=sm[0]
                #print "New balance: ",new_balance
                #new_balance=0
                
                if not fo_start:
                    #print "Setting"
                    sgn=self.drive.move_fw(self.duty_straight,self.balance_calibration+new_balance)
                    self.state.set_odo_signs(sgn)
            
            
            
            #print "yl",yl
            
            
            #Co simulation for damping out
            cs=self.co_simulation_straight([yl[0].tolist()[0][0],ydl[0].tolist()[0][0]])
            #cs=yl[0]
            
            x_target=p[2]
            if math.fabs(cs)>=math.fabs(x_target)-self.eps_lin and not fo_start:
                if False:
                    print "Breaking"
                    print cs
                    print "##############################"
                self.drive.break_move()            
                fo_start=t
                    
            if fo_start:
                #print t-fo_start
                if t-fo_start>=self.fo_time:
                    time.sleep(0.5)
                    break
            

            time.sleep(0.01)
          
        print "Position after straight\n",y,"\n##########"
        
    def tr(self,p,ref):
        
        fo_start=False

        while True:
            #Get state / Position /Velocity in global and local coordinates
            #print self.state.t_label
            t=self.state.t[0]
            y=np.matrix(self.state.y[0:3]).transpose()
            yd=np.matrix(self.state.y[3:6]).transpose()
            yl=self.global_to_local(ref,y)
            ydl=self.global_to_local(ref,yd,"rotate")
                        
            #Ultrasonic
            us_arr=[float(i) for j in self.us.distance for i in j]
            self.us_data.append({"state":y,"us":us_arr})
            
            if not fo_start:
                sgn=self.drive.turn_right(self.duty_turn)
                self.state.set_odo_signs(sgn)
                
            #Co simulation for damping out
            #print "yd",yd
            #print "ydl",ydl
            cs=self.co_simulation_turn([yl[2].tolist()[0][0],ydl[2].tolist()[0][0]])
            
            #print "y",y
            #print "yl",yl
            #print "cs ",cs
            
            x_target=p[2]
            if math.fabs(cs)>=math.fabs(x_target)-self.eps_turn and not fo_start:
                if False:
                    print "Breaking"
                    print cs
                    print "##############################"
                self.drive.break_move()            
                
                fo_start=t
                    
            if fo_start:
                #print t-fo_start
                if t-fo_start>=self.fo_time:
                    time.sleep(0.5)
                    break
            

            time.sleep(0.01)
        print "Position after turn\n",y,"\n##########"
        

s=sequence()
    
s.execute_sequence([
                    ["fw",False,0.70],
                    ["tr",False,math.radians(90)],
                    ["fw",False,0.70],
                    ["tr",False,math.radians(90)],
                    ["fw",False,0.70],
                    ["tr",False,math.radians(90)],
                    ["fw",False,0.70],
                    ["tr",False,math.radians(90)]
                    ])

#print s.us_data

with open("map_hash.dat", 'wb') as f:
    pickle.dump(s.us_data, f)
    
print "Executed successfully"
sys.exit()